"""Module in charge of detecting mock usage in expanded source code."""


__all__ = [
    "collect_mocked_functions",
    "IncludeDirective",
    "MockedFunction",
    "Token",
    "ForgivingDeclarationParser",
]


import os
import sys
import re
from collections import defaultdict
from typing import NamedTuple, List, Tuple

from pycparser import c_ast as node
from pycparser.c_parser import CParser
from pycparser.plyparser import ParseError


GETTER_REGEX = re.compile(
    r"(_narmock_state_type_for_[A-Za-z0-9_]+\s*\*\s*)?\b"
    + fr"_narmock_get_mock_for_([A-Za-z0-9_]+)\s*\("
)


def collect_mocked_functions(expanded_source_code):
    """Yield all the mocked functions used in the expanded source code."""

    functions = set()

    for match in GETTER_REGEX.finditer(expanded_source_code):
        return_type, function_name = match.groups()
        if not return_type:
            functions.add(function_name)

    yield from ForgivingDeclarationParser(expanded_source_code, functions)

    if functions:
        for function in functions:
            print("error:", f"'{function}' undeclared", file=sys.stderr)
        sys.exit(1)


def rename_arguments(function_declaration):
    if not function_declaration.type.args:
        return function_declaration

    for i, param in enumerate(function_declaration.type.args.params):
        param_type = param.type

        if (
            not param.name
            and isinstance(param_type.type, node.IdentifierType)
            and param_type.type.names == ["void"]
        ):
            continue

        param.name = f"arg{i + 1}"

        while not isinstance(param_type, node.TypeDecl):
            param_type = param_type.type

        param_type.declname = param.name

    return function_declaration


class IncludeDirective(NamedTuple):
    path: str
    system: bool

    @classmethod
    def from_source_context(cls, source_context):
        if not source_context:
            return None

        for filename in source_context:
            if not os.path.isabs(filename):
                continue

            dirname, basename = os.path.split(filename)
            fullname = basename

            while True:
                dirname, basename = os.path.split(dirname)

                if "include" in basename.lower():
                    return cls(fullname, True)

                if not basename:
                    break

                fullname = os.path.join(basename, fullname)

        return cls(os.path.abspath(source_context[-1]), False)

    def __str__(self):
        return "#include " + (f"<{self.path}>" if self.system else f'"{self.path}"')


class MockedFunction(NamedTuple):
    name: str
    declaration: str
    include: IncludeDirective


class Token(NamedTuple):
    type: str
    value: str
    span: Tuple[int, int]

    def is_punctuation(self, *symbols):
        return self.type == "PUNCTUATION" and self.value in symbols

    def is_keyword(self, *keywords):
        return self.type == "KEYWORD" and self.value in keywords

    @property
    def is_prefix(self):
        return self.type == "KEYWORD" and self.value in (
            "typedef",
            "extern",
            "static",
            "auto",
            "register",
            "__extension__",
        )


class ForgivingDeclarationParser:
    linemarker = re.compile(r'^# \d+ "((?:\\.|[^\\"])*)"((?: [1234])*)$')

    tokens = {
        "LINEMARKER": r"^#.*$",
        "KEYWORD": (
            "\\b(?:auto|break|case|char|const|continue|default|do|double|else|enum|extern|float"
            "|for|goto|if|int|long|register|return|short|signed|sizeof|static|struct|switch"
            "|typedef|union|unsigned|void|volatile|while|__extension__)\\b"
        ),
        "IDENTIFIER": r"\b[a-zA-Z_](?:[a-zA-Z_0-9])*\b",
        "CHARACTER": r"L?'(?:\\.|[^\\'])+'",
        "STRING": r'L?"(?:\\.|[^\\"])*"',
        "INTEGER": r"(?:0[xX][a-fA-F0-9]+|[0-9]+)[uUlL]*",
        "FLOAT": (
            r"(?:[0-9]+[Ee][+-]?[0-9]+|[0-9]*\.[0-9]+(?:[Ee][+-]?[0-9]+)?|[0-9]+\.[0-9]*(?:[Ee][+-]?[0-9]+)?)[fFlL]?"
        ),
        "PUNCTUATION": (
            r"\.\.\.|>>=|<<=|\+=|-=|\*=|/=|%=|&=|\^=|\|=|>>|<<|\+\+|--|->|&&|\|\||<=|>=|"
            r"==|!=|;|\{|\}|,|:|=|\(|\)|\[|\]|\.|&|!|~|-|\+|\*|/|%|<|>|\^|\||\?"
        ),
        "SPACE": r"[ \t\v\n\f]*",
        "IGNORE": r".+?",
    }

    ignored_tokens = "SPACE", "IGNORE"

    regex = re.compile(
        "|".join(f"(?P<{token}>{pattern})" for token, pattern in tokens.items()),
        flags=re.MULTILINE,
    )

    def __init__(self, source_code, functions=None):
        self.source_code = source_code
        self.functions = functions
        self.token_stream = self.tokenize(source_code)
        self.previous = None
        self.current = None

        self.bracket_stack = []
        self.source_context = []
        self.typedefs = ["typedef int __builtin_va_list;"]

        self.cparser = CParser()

    @classmethod
    def tokenize(cls, source_code):
        for match in cls.regex.finditer(source_code):
            if match.lastgroup not in cls.ignored_tokens:
                yield Token(match.lastgroup, match.group().strip(), match.span())

    def __iter__(self):
        while self.next():
            if self.current.is_keyword("typedef"):
                self.parse_typedef()

            function = self.parse_function_declaration()

            if function is not None:
                yield function

            if self.functions is not None and not self.functions:
                break

            while self.current and not (
                self.current.is_punctuation(";", "}") and not self.bracket_stack
            ):
                self.next()

    def next(self):
        self.previous = self.current
        self.current = next(self.token_stream, None)

        if not self.current:
            return None

        if self.current.type == "PUNCTUATION":
            if self.current.value in "({[":
                self.bracket_stack.append(")}]"["({[".index(self.current.value)])
            elif self.bracket_stack and self.current.value == self.bracket_stack[-1]:
                self.bracket_stack.pop()

        elif self.current.type == "LINEMARKER":
            filename, flags = self.linemarker.match(self.current.value).groups()

            if "1" in flags:
                self.source_context.append(filename)
            elif "2" in flags:
                self.source_context.pop()

            try:
                self.source_context[-1] = filename
            except IndexError:
                self.source_context.append(filename)

            self.next()

        return self.current

    def parse_typedef(self):
        start_index = self.current.span[0]

        while self.current and not (
            self.current.is_punctuation(";") and not self.bracket_stack
        ):
            self.next()

        self.typedefs.append(self.source_code[start_index : self.current.span[1]])

    def parse_function_declaration(self):
        if self.bracket_stack:
            return None

        while self.current and self.current.is_prefix:
            self.next()

        start_index = self.current.span[0]
        return_type = []

        while (
            self.current
            and not self.current.is_punctuation("(")
            or self.next()
            and self.current.is_punctuation("*")
        ):
            if not self.bracket_stack and self.current.is_punctuation(";"):
                return None

            return_type.append(self.current.value)
            self.next()

        if not return_type:
            return None

        func_name = return_type.pop()

        if self.functions is not None and func_name not in self.functions:
            return None

        while (
            self.current
            and self.bracket_stack
            or self.next()
            and self.current.is_punctuation("(")
        ):
            self.next()

        signature = self.source_code[start_index : self.previous.span[1]] + ";"
        code = "\n".join(self.typedefs) + "\n" + signature

        try:
            file_ast = self.cparser.parse(code)
        except ParseError:
            return None
        else:
            if self.functions is not None:
                self.functions.remove(func_name)

            return MockedFunction(
                func_name,
                rename_arguments(file_ast.ext[-1]),
                IncludeDirective.from_source_context(self.source_context),
            )



"""Module in charge of generating mocks."""


__all__ = ["GeneratedMock", "FileGenerator"]


import os
import re
from copy import deepcopy

from jinja2 import Environment, PackageLoader
from pycparser import c_ast as node
from pycparser.c_generator import CGenerator

from . import __version__


def decl(name, type):
    return node.Decl(name, [], [], [], type, None, None)


def function_ptr_decl(name, return_type, parameters):
    return decl(
        name, node.PtrDecl([], node.FuncDecl(node.ParamList(parameters), return_type))
    )


def rename_return_type(return_type, name):
    return_type = deepcopy(return_type)
    type_decl = return_type

    while not isinstance(type_decl, node.TypeDecl):
        type_decl = type_decl.type

    type_decl.declname = name
    return return_type


def get_guard_name(filename):
    slug = re.sub(r"[^a-zA-Z0-9]", "_", os.path.normpath(os.path.relpath(filename)))
    return re.sub(r"_+", "_", slug).upper().strip("_")


def generate_includes(system_includes, local_includes, directory):
    return "\n\n".join(
        includes
        for includes in (
            "\n".join(f"#include <{path}>" for path in sorted(system_includes)),
            "\n".join(
                f'#include "{os.path.relpath(path, directory)}"'
                for path in sorted(local_includes)
            ),
        )
        if includes
    )


class GeneratedMock:
    DECL_MARKER = "// NARMOCK_DECLARATION"
    IMPL_MARKER = "// NARMOCK_IMPLEMENTATION"
    FLAGS_MARKER = "// NARMOCK_LINKER_FLAGS"

    FLAGS_REGEX = re.compile(fr"^{FLAGS_MARKER}\s+(.+)$", re.MULTILINE)

    @classmethod
    def extract_flags(cls, mock_declarations):
        return cls.FLAGS_REGEX.findall(mock_declarations)

    def __init__(self, function):
        self.function = function

        self.func_name = function.name

        self.wrapped_func = f"__wrap_{self.func_name}"
        self.real_func = f"__real_{self.func_name}"
        self.linker_flags = f"-Wl,--wrap={self.func_name}"

        self.state_name = f"_narmock_state_for_{self.func_name}"
        self.state_type = f"_narmock_state_type_for_{self.func_name}"
        self.private_state_type = f"_narmock_private_state_type_for_{self.func_name}"

        self.func_decl = self.function.declaration.type
        self.func_params = self.func_decl.args.params if self.func_decl.args else []
        self.forward_args = ", ".join(
            param.name for param in self.func_params if param.name
        )

        return_type = self.func_decl.type
        self.return_value = (
            None
            if isinstance(return_type, node.TypeDecl)
            and isinstance(return_type.type, node.IdentifierType)
            and return_type.type.names[0] == "void"
            else "return_value"
        )

        self.return_value_decl = decl(
            self.return_value, rename_return_type(return_type, self.return_value)
        )
        self.implementation_decl = function_ptr_decl(
            "implementation",
            rename_return_type(return_type, "implementation"),
            self.func_params,
        )

        self.mock_return_decl = self.state_function(
            "mock_return", [self.return_value_decl]
        )
        self.mock_implementation_decl = self.state_function(
            "mock_implementation", [self.implementation_decl]
        )
        self.disable_mock_decl = self.state_function(
            "disable_mock",
            [decl(None, node.TypeDecl(None, [], node.IdentifierType(["void"])))],
        )

        self.real_decl = self.rename_function(self.real_func)
        self.wrapped_decl = self.rename_function(self.wrapped_func)

    def state_function(self, name, parameters):
        return function_ptr_decl(
            name,
            node.PtrDecl(
                [], node.TypeDecl(name, [], node.IdentifierType([self.state_type]))
            ),
            parameters,
        )

    def rename_function(self, name):
        return decl(
            name,
            node.FuncDecl(
                self.func_decl.args, rename_return_type(self.func_decl.type, name)
            ),
        )


class FileGenerator:
    SOURCE_FILE = "__mocks__.c"
    HEADER_FILE = "__mocks__.h"

    def __init__(self):
        self.code_generator = CGenerator()

        self.jinja_env = Environment(
            loader=PackageLoader("narmock", "templates"),
            trim_blocks=True,
            lstrip_blocks=True,
        )
        self.jinja_env.filters["render"] = self.code_generator.visit

        self.source_template = self.jinja_env.get_template(f"{self.SOURCE_FILE}.jinja2")
        self.header_template = self.jinja_env.get_template(f"{self.HEADER_FILE}.jinja2")

        self.mocks = []
        self.system_includes = set()
        self.local_includes = set()

    def add_mock(self, mocked_function):
        self.mocks.append(GeneratedMock(mocked_function))

        if mocked_function.include:
            if mocked_function.include.system:
                self.system_includes.add(mocked_function.include.path)
            else:
                self.local_includes.add(mocked_function.include.path)

    def write_to_directory(self, directory):
        source_filename = os.path.join(directory, self.SOURCE_FILE)
        header_filename = os.path.join(directory, self.HEADER_FILE)

        mocks = list(sorted(self.mocks, key=lambda m: m.func_name))

        source_code = self.source_template.render(
            narmock_version=__version__,
            includes=generate_includes([], [header_filename], directory),
            mocks=mocks,
        )

        header_code = self.header_template.render(
            narmock_version=__version__,
            guard_name=get_guard_name(header_filename),
            includes=generate_includes(
                self.system_includes, self.local_includes, directory
            ),
            mocks=mocks,
        )

        with open(source_filename, "w") as source_file:
            source_file.write(source_code.strip() + "\n")

        with open(header_filename, "w") as header_file:
            header_file.write(header_code.strip() + "\n")

    @classmethod
    def read_declarations(cls, directory):
        with open(os.path.join(directory, cls.HEADER_FILE)) as header_file:
            return header_file.read()



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['narmock', 'narmock.templates']

package_data = \
{'': ['*']}

install_requires = \
['jinja2>=2.10,<3.0', 'pycparser>=2.19,<3.0']

entry_points = \
{'console_scripts': ['narmock = narmock.cli:main']}

setup_kwargs = {
    'name': 'narmock',
    'version': '0.2.0',
    'description': 'A minimal mocking utility for C projects.',
    'long_description': '# 🎣 narmock\n\n[![Build Status](https://travis-ci.com/vberlier/narmock.svg?branch=master)](https://travis-ci.com/vberlier/narmock)\n[![PyPI](https://img.shields.io/pypi/v/narmock.svg)](https://pypi.org/project/narmock/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/narmock.svg)](https://pypi.org/project/narmock/)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n\n> A minimal mocking utility for C projects.\n\n**🚧 Work in progress 🚧**\n\nNarmock finds the functions mocked in your tests and generates mocks with a slick API.\n\n```c\n#include <time.h>\n\n#include "__mocks__.h"\n#include "narwhal.h"\n\nTEST(example)\n{\n    MOCK(time)->mock_return(42);\n    ASSERT_EQ(time(NULL), 42);\n}\n```\n\n> This example is a test written with [Narwhal](https://github.com/vberlier/narwhal) but Narmock can be used with other test frameworks and anywhere in regular source code.\n\n## Installation\n\nThe package can be installed with `pip`.\n\n```bash\n$ pip install narmock\n```\n\n## Getting started\n\nThe command-line utility provides two essential commands that should make it possible to integrate Narmock in any kind of build system.\n\n```\nusage: narmock [-h] (-g [<code>] | -f) [-d <directory>]\n\nA minimal mocking utility for C projects.\n\noptional arguments:\n  -h, --help      show this help message and exit\n  -g [<code>]     generate mocks\n  -f              output linker flags\n  -d <directory>  mocks directory\n```\n\n> Check out the [basic example](https://github.com/vberlier/narmock/tree/master/examples/basic) for a simple Makefile that integrates both [Narwhal](https://github.com/vberlier/narwhal) and Narmock.\n\n### Generating mocks\n\nThe `narmock -g` command finds the functions mocked in your code and generates a `__mocks__.c` file and a `__mocks__.h` file that respectively define and declare all the required mocks.\n\n```bash\n$ gcc -E *.c | narmock -g\n```\n\nNarmock requires source code to be expanded by the preprocessor. You can directly pipe the output of `gcc -E` to the command-line utility.\n\nBy default, `__mocks__.c` and `__mocks__.h` will be created in the current directory. You can specify a different output directory with the `-d` option.\n\n```bash\n$ gcc -E tests/*.c | narmock -g -d tests\n```\n\n### Retrieving linker flags\n\nThe `narmock -f` command reads the generated `__mocks__.h` file and outputs the necessary linker flags for linking all your source files together.\n\n```bash\n$ gcc $(narmock -f) *.c\n```\n\nBy default, the command looks for `__mocks__.h` in the current directory. You can specify a different directory with the `-d` option.\n\n```bash\n$ gcc $(narmock -f -d tests) tests/*.c\n```\n\n## Mock API\n\nThe `MOCK` macro returns a pointer to the mock API of a given function.\n\n```c\nMOCK(time);\n```\n\n### Mock return\n\nYou can make a function return a specific value without calling its original implementation.\n\n```c\nMOCK(time)->mock_return(42);\n\nprintf("%ld\\n", time(NULL));  // Outputs 42\n```\n\n### Mock implementation\n\nYou can switch the implementation of a function.\n\n```c\ntime_t time_stub(time_t *timer)\n{\n    return 42;\n}\n\nMOCK(time)->mock_implementation(time_stub);\n\nprintf("%ld\\n", time(NULL));  // Outputs 42\n```\n\n### Disable mock\n\nYou can disable the mock and make the function call its original implementation.\n\n```c\nMOCK(time)->disable_mock();\n\nprintf("%ld\\n", time(NULL));  // Outputs the current time\n```\n\n## Contributing\n\nContributions are welcome. Feel free to open issues and suggest improvements. This project uses [poetry](https://poetry.eustace.io/) so you\'ll need to install it first if you want to be able to work with the project locally.\n\n```bash\n$ curl -sSL https://raw.githubusercontent.com/sdispater/poetry/master/get-poetry.py | python\n```\n\nYou should now be able to install the required dependencies.\n\n```bash\n$ poetry install\n```\n\nThe code follows the [black](https://github.com/ambv/black) code style.\n\n```bash\n$ poetry run black narmock\n```\n\nYou can run the tests with `poetry run make -C tests`. The test suite is built with [Narwhal](https://github.com/vberlier/narwhal).\n\n```bash\n$ poetry run make -C tests\n```\n\n---\n\nLicense - [MIT](https://github.com/vberlier/narmock/blob/master/LICENSE)\n',
    'author': 'Valentin Berlier',
    'author_email': 'berlier.v@gmail.com',
    'url': 'https://github.com/vberlier/narmock',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



"""Commande-line interface."""


__all__ = ["main"]


import sys
from argparse import ArgumentParser, FileType

from .api import generate_mocks, collect_linker_flags


parser = ArgumentParser(
    prog="narmock", description="A minimal mocking utility for C projects."
)

group = parser.add_mutually_exclusive_group(required=True)
group.add_argument(
    "-g",
    metavar="<code>",
    nargs="?",
    type=FileType("r"),
    const=sys.stdin,
    help="generate mocks",
)
group.add_argument("-f", action="store_true", help="output linker flags")

parser.add_argument("-d", metavar="<directory>", default=".", help="mocks directory")


def main():
    args = parser.parse_args()

    if args.g:
        generate_mocks(expanded_code=args.g.read(), directory=args.d)
    elif args.f:
        print(" ".join(collect_linker_flags(directory=args.d)).strip())
